"use strict";
//https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest
//http://stackoverflow.com/questions/2130452/how-to-efficiently-append-content-with-jquery
//http://jscrollpane.kelvinluck.com/
//http://jscrollpane.kelvinluck.com/settings.html
//http://jscrollpane.kelvinluck.com/api.html
var T = function() {				// module
    var transcodersXML = {};
    var transcoders = {};
    var api;		// must be set to Kevin Luck's scrolling code
    var asisStart = '[';
    var asisEnd = ']';
    var blockSize = 8192;
    var currentTargetHighlighted = null;
    var prevMatchedLineText = null;
    var prevMatchedLineObj = null;
    var currentAnswer = -1;
    var answers = [];
    var transcodeBuffer = "";
    var currentInputTranscoderSLP = null;
    var currentOutputTranscoderSLP = null;
    var currentOtherToSLPTranscoder = null;
    var currentSLPToOtherTranscoder = null;
    var lastKeyDeletion = false;
    var transcriptionOptions = {};
    var inputUsingSLP = true;
    var lastValue = "";
    var virama = {
        "\u094d": true, // devanagari
        "\u09cd": true, // bengali
        "\u0a4d": true, // gurmukhi
        "\u0acd": true, // gujarati
        "\u0c4d": true, // telugu
        "\u0ccd": true, // kannada
        "\u0d4d": true		// malayalam
    };

    var slpBuffer = "";
    var otherBuffer = "";

    var SLP_vowel = 1;
    var SLP_consonent = 2;
    var SLP_accent = 4;
    var SLP_whitespace = 8;
    var SLP_atom = 16;

    var slpCategories = {
        "a": SLP_vowel,
        "A": SLP_vowel,
        "i": SLP_vowel,
        "I": SLP_vowel,
        "u": SLP_vowel,
        "U": SLP_vowel,
        "f": SLP_vowel,
        "F": SLP_vowel,
        "x": SLP_vowel,
        "X": SLP_vowel,
        "e": SLP_vowel,
        "E": SLP_vowel,
        "o": SLP_vowel,
        "O": SLP_vowel,
        "H": SLP_vowel,
        "m": SLP_consonent,
        "M": SLP_vowel,
        "k": SLP_consonent,
        "K": SLP_consonent,
        "g": SLP_consonent,
        "G": SLP_consonent,
        "N": SLP_consonent,
        "c": SLP_consonent,
        "C": SLP_consonent,
        "j": SLP_consonent,
        "J": SLP_consonent,
        "Y": SLP_consonent,
        "w": SLP_consonent,
        "W": SLP_consonent,
        "q": SLP_consonent,
        "Q": SLP_consonent,
        "R": SLP_consonent,
        "t": SLP_consonent,
        "T": SLP_consonent,
        "d": SLP_consonent,
        "D": SLP_consonent,
        "n": SLP_consonent,
        "p": SLP_consonent,
        "P": SLP_consonent,
        "b": SLP_consonent,
        "B": SLP_consonent,
        "y": SLP_consonent,
        "r": SLP_consonent,
        "l": SLP_consonent,
        "v": SLP_consonent,
        "s": SLP_consonent,
        "S": SLP_consonent,
        "z": SLP_consonent,
        "h": SLP_consonent,
        "L": SLP_consonent,
        "|": SLP_consonent,
        "'": SLP_atom,
        ";": SLP_atom,
        "-": SLP_atom,
        "~": SLP_accent,
        "\\": SLP_accent,
        " ": SLP_whitespace,
        "\t": SLP_whitespace,
        "\r": SLP_whitespace,
        "\n": SLP_whitespace,
        ".": SLP_atom,
        "0": SLP_atom,
        "1": SLP_atom,
        "2": SLP_atom,
        "3": SLP_atom,
        "4": SLP_atom,
        "5": SLP_atom,
        "6": SLP_atom,
        "7": SLP_atom,
        "8": SLP_atom,
        "9": SLP_atom
    };

    var inputIsSLP = function() {
        return inputUsingSLP;
    };

    var getOtherBuffer = function() {
        return otherBuffer;
    };

    var addHandlersForDragging = function(id) {
        $(id).on('mousedown', function(e) {
            var node = $(this);
            var position = node.offset();
            var initialized = {
                x: position.left - e.pageX,
                y: position.top - e.pageY
            };
            var handlers = {
                mousemove: function(e) {
                    currentBox.x = initialized.x + e.pageX;
                    currentBox.y = initialized.y + e.pageY;
                    node.css({
                        left: (currentBox.x) + 'px',
                        top: (currentBox.y) + 'px'
                    });
                },
                mouseup: function(e) {
                    $(this).off(handlers);
                }
            };
            $(document).on(handlers);
        });		// end on handler
    };

    var getEchoTypingUsingDisplay = function() {
        var echoTypingUsingDisplay;

        echoTypingUsingDisplay = localStorage.getItem("echoTypingUsingDisplay");
        if (!echoTypingUsingDisplay) {
            echoTypingUsingDisplay = true;
            setEchoTypingUsingDisplay(true);
        } else if (echoTypingUsingDisplay === "true") {
            echoTypingUsingDisplay = true;
        } else {
            echoTypingUsingDisplay = false;
        }

        return echoTypingUsingDisplay;
    };

    var setEchoTypingUsingDisplay = function(b) {
        localStorage.setItem("echoTypingUsingDisplay", b);
    };


    var addHandlersForTyping = function(id, callback) {
        //http://stackoverflow.com/questions/5571514/convert-to-uppercase-as-user-types-using-javascript

        var echoTypingUsingDisplay;

        echoTypingUsingDisplay = getEchoTypingUsingDisplay();

        if (!inputIsSLP() || echoTypingUsingDisplay) {
            return;		// input is not SLP, do not intercept keystrokes
        }

        $(id).on('input', function(evt) {
            var char, input, start;

            input = $(this);

            transcriptionOptions = {};

            start = input[0].selectionStart;
            if (!lastKeyDeletion) {
                char = input[0].value.charAt(start - 1);
                if (start === input[0].value.length) {   // new character appended at end
                    slpBuffer += char;
                } else {
                    transcriptionOptions.insertionPoint = insertInSLPBuffer(input[0].value, start);
                }
            }
            transcriptionOptions.start = start;

            $(this).val(function(_, val) {
                var len, result, n;
                result = {incrementCursor: 0, text: ""};
                if (lastKeyDeletion) {
                    if (slpBuffer.length > 0) {
                        result = deleteFromSLPBuffer(input[0].value, this.selectionEnd);   // need to make slpBuffer consistent with value shown in textarea
                    }
                    if (result) {
                        if (result.incrementCursor !== 0) {
                            n = transcriptionOptions.start + result.incrementCursor;
                            if (n >= 0) {
                                transcriptionOptions.start = n;
                            }
                        }
                        return result.text;
                    } else {
                        return this.value;      // leave as is.
                    }
                } else {
                    result = transcodeText([slpBuffer])[0];
                    if (transcriptionOptions.insertionPoint) {
                        transcriptionOptions.start = transcriptionOptions.insertionPoint;
                    } else {
                        transcriptionOptions.start = result.length;
                    }
                    return result;
                }
            });	// end on val handler

            input[0].selectionStart = input[0].selectionEnd = transcriptionOptions.start;
            lastValue = input[0].value;
            callback(slpBuffer);
        }); // end on input handler


        $(id).keydown(function(evt) {
            lastKeyDeletion = evt.keyCode === 8;
        });
    };	// end addHandlersForTyping

    var insertInSLPBuffer = function(value, start) {
        var len, i, prefixDeva, prefixSLP, prefixCandidate, n, len, ch, buffer;
        // value includes the character just typed (which is not devanagari yet)
        // start will be the index of the character following the character just typed.
        n = 0;      // new character will be inserted before slpBuffer[n]
        len = slpBuffer.length;
        if (start > 1) {
            prefixDeva = value.substring(0, start - 1);     // will sequence through slpBuffer until it equals this value
            while (n < len) {
                n++;
                prefixSLP = slpBuffer.substr(0, n);
                prefixCandidate = transcodeText([prefixSLP]);
                if (prefixCandidate[0] === prefixDeva) {
                    break;
                }
            }
        }
        ch = value.charAt(start - 1);
        buffer = slpBuffer.substr(0, n) + ch;
        len = transcodeText([buffer])[0].length;         // cursor goes after inserted character
        slpBuffer = buffer + slpBuffer.substring(n);

        return len;
    };


    // Detect this situation (deleting blank from conjunct)
    // before: k, k,
    // after: kk,
    var spaceDeletionCreatesConjunct = function(value) {
        var len, i, b;
        b = false;
        len = lastValue.length;
        for (i = 1; i < len; i++) {
            if (value.charAt(i) !== lastValue.charAt(i)) {
                b = lastValue.charAt(i) === ' ' && virama[lastValue.charCodeAt(i - 1)];
                break;
            }
        }

        if (b) {
            return i;
        } else {
            return -1;
        }
    };


    var deleteFromSLPBuffer = function(value, start) {
        var len, i, n, typeDeleted, typePrevious, incrementCursor, modify, blankIdx, ch;
        // value does not include the deleted character
        // start will be the index of the character following the deleted character

        blankIdx = spaceDeletionCreatesConjunct(value);
        if (blankIdx !== -1) {
            n = locateDeletion(blankIdx, lastValue);
            removeCharacter(n);
            return null;
        }

        n = locateDeletion(start, value);       // n is index of character to be deleted from slpBuffer

        // now determine  what has to be deleted in slpBuffer

        modify = true;
        incrementCursor = 0;
        if (n >= slpBuffer.length) {         // a virama was deleted but not the character
            removeCharacter(slpBuffer.length - 1);  // the virama has to be the last character otherwise cursor could not have been positioned there.
        } else {
            typeDeleted = slpCategories[slpBuffer.charAt(n)];
            if (n !== 0) {
                typePrevious = slpCategories[slpBuffer.charAt(n - 1)];
            } else {
                typePrevious = 0;
            }

            if (typeDeleted === SLP_consonent) {
                if (n !== slpBuffer.length && slpBuffer.charAt(n + 1) === 'a') {   // if 'a' was removed
                    removeCharacter(n + 1);
                    incrementCursor = 2;        // put cursor after virama
                } else {
                    removeCharacter(n);
                }
            } else if (typePrevious === SLP_consonent) {
                if (typeDeleted & (SLP_vowel | SLP_accent)) {        // attached vowel or accent deleted
                    removeCharacter(n);
                } else {            // virama was deleted
                    removeCharacter(n - 1);
                    incrementCursor = -1;   // since the character is now gone
                }
            } else {
                removeCharacter(n);
                modify = false;
            }
        }

        if (modify) {
            return {incrementCursor: incrementCursor, text: transcodeText([slpBuffer])[0]};
        } else {
            return null;
        }
    };		// end deleteFromSLPBuffer


    var locateDeletion = function(len, value) {
        var i, n, ch;

        n = 0;
        for (i = 0; i < len; i++) {
            ch = value.charCodeAt(i);
            if (!virama[ch]) {      // don't count viramas since slpBuffer doesn't have them
                ch = slpBuffer.charAt(n);  // preceding char
                n++;
                if (n !== slpBuffer.length && slpCategories[ch] === SLP_consonent && slpBuffer.charAt(n) === "a") {
                    n++;        // since attached a has no unicode character.
                }
            }
        }

        return n;
    };

    var removeCharacter = function(n) {
        var buffer;

        buffer = slpBuffer.substring(0, n);     // buffer will not have nth character
        if (n !== slpBuffer.length) {
            buffer += slpBuffer.substring(n + 1);
        }

        slpBuffer = buffer;
    };


    var removeDeletion = function(n) {
        var ch, start, end, len, buffer;
        ch = slpBuffer.charAt(n);
        start = n;
        end = n;
        len = slpBuffer.length;
        if (slpCategories[ch] & (SLP_whitespace | SLP_atom)) {
            // delete single character
            end++;
        } else {
            // first skip consonents
            while (slpCategories[ch] === SLP_consonent) {
                end++;
                if (end !== len) {
                    ch = slpBuffer.charAt(end);
                } else {
                    break;
                }
            }

            // now skip vowels, if any
            while (slpCategories[ch] & (SLP_vowel | SLP_accent)) {
                end++;
                if (end !== len) {
                    ch = slpBuffer.charAt(end);
                } else {
                    break;
                }
            }
        }

        buffer = slpBuffer.substring(0, start);
        if (end !== len) {
            buffer += slpBuffer.substring(end);
        }

        slpBuffer = buffer;
    };


    var positionFloatingTextarea = function(id, location) {
        $(id).css('left', location.x)
                .css('top', location.y);

        if (location.width) {
            $(id).width(location.width);
        }
    };

    var setValue = function(id, obj) {
        if (typeof obj.value !== 'undefined') {
            $(id).val(obj.value);
        }

        if (typeof obj.slp !== 'undefined') {
            slpBuffer = obj.slp;
        }
    };

    var getSLPBuffer = function() {
        return slpBuffer;
    };

    var setSLPBuffer = function(slp) {
        slpBuffer = slp;
    };

    var getXMLHttpRequest = function() {
        var xhp = null;
        if (window.XMLHttpRequest) {
            xhp = new window.XMLHttpRequest();
        }
        else {
            try {
                xhp = new ActiveXObject("MSXML2.XMLHTTP");
            }
            catch (ex) {
                alert(ex);
            }
        }
        return xhp;
    };

    var init = function() {
        document.getElementById("query").onkeydown = function(event) {
            return getKeyStroke(event);
        };
        document.getElementById("query").onkeyup = function(event) {
            return transcode(event);
        };
    };


    var getItem = function(key, dflt) {
        var value;

        value = localStorage.getItem(key);
        if (value) {
            return value;
        } else {
            return dflt;
        }
    };

    var loadTranscoders = function() {
        var xhp, url, elText, blocks, parser, xmlDoc, inputEncoding, outputEncoding, a, len, i;

        inputEncoding = getItem("inputEncoding", "slp1");
        outputEncoding = getItem("outputEncoding", "roman");

        a = [];
        a.push("slp1_" + outputEncoding);
        a.push(outputEncoding + "_slp1");
        if (inputEncoding !== "slp1") {				// if we also need to map back and forth between input encoding and slp1
            a.push("slp1_" + inputEncoding);
            a.push(inputEncoding + "_slp1");
            inputUsingSLP = false;
        } else {
            inputUsingSLP = true;
        }

        len = a.length;
        for (i = 0; i < len; i++) {
            if (!transcoders[a[i]]) {
                parser = new DOMParser();
                xmlDoc = parser.parseFromString(transcodersXML[a[i]], "text/xml");
                transcoders[a[i]] = convertXML(xmlDoc);
            }
        }

        currentInputTranscoderSLP = transcoders[a[0]];		// slp1 --> output encoding
        currentOutputTranscoderSLP = transcoders[a[1]];		// output encoding --> slp1
        currentSLPToOtherTranscoder = transcoders[a[2]];	// may be null
        currentOtherToSLPTranscoder = transcoders[a[3]];	// may be null
    };


    var displayText = function(blocks) {
        var len, i;

        len = blocks.length;
        for (i = 0; i < len; i++) {
            api.getContentPane().append(blocks[i]);
        }
        api.reinitialise();
    };

    var convertXML = function(xml) {
        var fsm, edges, edge, len, i, j, e, t, el, ctx, ch, list, h, s, cl, transcoder;

        transcoder = {};

        fsm = xml.getElementsByTagName("fsm")[0];

        el = fsm.attributes.getNamedItem("inputType");
        if (el) {
            transcoder.inputType = el.value;		// note: was warned that el.nodeValue was deprecated
        } else {
            transcoder.inputType = null;
        }

        el = fsm.attributes.getNamedItem("outputType");
        if (el) {
            transcoder.outputType = el.value;
        } else {
            transcoder.outputType = null;
        }

        el = fsm.attributes.getNamedItem("version");
        if (el) {
            transcoder.version = el.value;
        } else {
            transcoder.version = null;
        }

        el = fsm.attributes.getNamedItem("capitalsSupported");
        if (el) {
            transcoder.capitalsSupported = el.value;
        } else {
            transcoder.capitalsSupported = null;
        }

        edges = fsm.getElementsByTagName("edges")[0];
        edge = edges.getElementsByTagName("e");
        len = edge.length;
        transcoder.classes = {};
        transcoder.edges = [];
        transcoder.h = {};
        for (i = 0; i < len; i++) {
            e = edge[i];
            t = {};
            transcoder.edges.push(t);

            t.n = e.attributes.getNamedItem("n").value;

            el = e.getElementsByTagName("in")[0];
            if (el) {
                t.in = convertUnicode(el.textContent);
                ch = t.in.charAt(0);
                list = transcoder.h[ch];

                if (list) {
                    list.push(t);
                } else {
                    transcoder.h[ch] = [t];
                }
            } else {
                t.in = null;
            }

            el = e.getElementsByTagName("out")[0];
            if (el) {
                t.out = convertUnicode(el.textContent);
            } else {
                t.out = null;
            }

            el = e.getElementsByTagName("cl")[0];
            if (el) {
                t.cls = el.textContent;
                cl = t.cls.split(";");
                if (cl.length > 1) {
                    j = 0;
                }
                for (j = 0; j < cl.length; j++) {
                    h = transcoder.classes[cl[j]];
                    if (!h) {
                        h = {};
                        transcoder.classes[cl[j]] = h;
                    }
                    h[t.in] = true;
                    ;
                }
            } else {
                t.cls = null;
            }

            ctx = e.getElementsByTagName("ctx")[0];
            if (ctx) {
                transcoder.contextual = true;
                t.ctx = {};
                el = ctx.attributes.getNamedItem("card");
                if (el) {
                    t.ctx.option = el.value;
                } else {
                    t.ctx.option = "+";
                }

                el = ctx.attributes.getNamedItem("dir");
                if (el) {
                    t.ctx.leftToRight = el.value.toLowerCase() === "after";
                } else {
                    t.ctx.leftToRight = true;
                }

                el = ctx.attributes.getNamedItem("x");
                if (el) {
                    t.ctx.exists = true;
                    t.ctx.eolOverride = false;
                    list = el.value.toLowerCase().split("|");
                    for (j = 0; j < list.length; j++) {
                        s = list[j];
                        if (s === "y") {
                            t.ctx.exists = true;
                        } else if (s === "n") {
                            t.ctx.exists = false;
                        } else if (s === "eol") {
                            t.ctx.eolOverride = true;
                        }
                    }
                } else {
                    t.ctx.exists = false;
                }

                t.ctx.category = ctx.textContent.substring(1);     // get rid of $    
            } else {
                transcoder.contextual = false;
                t.ctx = null;
            }
        }

        return transcoder;
    };  // convertXML

    // convert unicode encoded text to raw text
    var convertUnicode = function(s) {
        return s.replace(
                /\\u([0-9a-f]{4})/g,
                function(whole, group1) {
                    return String.fromCharCode(parseInt(group1, 16));
                }
        );
    };

    var transcodeTextSafe = function(lines, options, offset) {
        try {
            return transcodeText(lines, options, offset);
        } catch (ex) {
            if (lines.length === 1) {
                alert("Error transcoding " + lines[0] + ": " + ex);
                return "";
            } else {
                alert("Error transcoding array: " + ex);
            }
        }
    };

    var transcodeSLPToOther = function(slp) {
        return currentSLPToOtherTranscoder.h[slp];
    };


    var transcodeOtherToSLP = function(other) {
        return transcodeText([other], {transcoder: currentOtherToSLPTranscoder})[0];
    };

    var transcodeText = function(lines, options, offset) {
        var len, i, ch, e, result, list, j, k, bestlen, bestidx, s, blocks, totalLength, line, numlines, lineno, buffer,
                generateParagraphs, firstline, lastline, inputTranscoder, parser, xmlDoc, asisOptions;

        numlines = lines.length;

        if (options) {
            generateParagraphs = options.generateParagraphs;
            firstline = options.firstline;
            lastline = options.lastline;
            inputTranscoder = options.transcoder;
            if (inputTranscoder === undefined) {
                inputTranscoder = currentInputTranscoderSLP;
            } else if (typeof inputTranscoder === "string") {
                if (transcoders[inputTranscoder] === null) {
                    parser = new DOMParser();
                    xmlDoc = parser.parseFromString(transcodersXML[inputTranscoder], "text/xml");
                    transcoders[inputTranscoder] = convertXML(xmlDoc);
                    inputTranscoder = transcoders[inputTranscoder];
                } else {
                    inputTranscoder = transcoders[inputTranscoder];
                }
            }
            // else inputTranscoder is ready to use
        } else {
            generateParagraphs = false;
            inputTranscoder = currentInputTranscoderSLP;
        }

        totalLength = 0;
        buffer = [];
        blocks = [];
        if (firstline === undefined) {
            firstline = 0;
            lastline = numlines;
        } else if (lastline === undefined) {
            lastline = numlines;
        }


        for (lineno = firstline; lineno < lastline; lineno++) {
            line = lines[lineno];
            if (offset) {
                line = line.substring(offset);
            }
            len = line.length;
            if (generateParagraphs) {
                result = "<p id='p" + lineno + "'>";
            } else {
                result = "";
            }
            for (i = 0; i < len; i++) {
                ch = line.charAt(i);
                if (ch === asisStart) {		// REB: In Java version, the value of the as is delimiter is an option
                    asisOptions = {line: line, offset: i};
                    getPassThrough(asisOptions);
                    i = asisOptions.offset;		// Will point to the end delimiter
                    result += asisOptions.passthrough;  // what is between the braces
                    continue;
                }

                list = inputTranscoder.h[ch];
                if (!list) {
                    continue;
                }
                if (list.length === 1) {
                    bestidx = 0;
                } else {
                    bestlen = 0;
                    bestidx = 0;
                    for (j = 0; j < list.length; j++) {
                        s = list[j].in;
                        k = s.length;
                        if (line.substr(i, k) === s) {
                            if (k > bestlen && contextCheck(inputTranscoder, line, i, list[j])) {
                                bestlen = k;
                                bestidx = j;
                            }
                        }
                    }
                }
                result += list[bestidx].out;
                if (bestlen > 1) {
                    i += bestlen - 1;		// prepare for i++ in for loop
                }
            }

            buffer.push(generateParagraphs ? (result + "</p>") : result);
            totalLength += result.length;
            if (totalLength > blockSize) {
                blocks.push(buffer.join("\n"));
                buffer.length = 0;
                totalLength = 0;
                result = "";
            }
        }  // end for each line
        s = buffer.join("\n");
        if (s.length > 0) {
            blocks.push(s);
        } else if (blocks.length === 0) {
            blocks = [""];  // must have been passed the empty string
        }
        return blocks;
    };  // end transcodeText

    var getPassThrough = function(options) {
        var line, offset, start, end, i, len;

        line = options.line;
        start = options.offset;
        if (line[start] === asisStart) {
            start++;
        }
        len = line.length;
        end = 0;
        for (i = start; i < len && !end; i++) {
            if (line[i] === asisEnd) {
                end = i;
            }
        }
        if (!end) {				// if ] not found, pass through everything after start
            end = line.length;
        }

        options.offset = end;
        options.passthrough = line.substring(start, end);
    };

    // check if text.substring(n) matches context (if any) 
    var contextCheck = function(transcoder, line, n, edge) {
        var found, idx, match, nextidx, currentidx, previdx, beforeidx;

        match = edge.in;
        idx = n + match.length;
        if (edge.ctx) {
            nextidx = idx;            // used by left to right
            currentidx = idx - match.length;       // used by right to left

            found = false;

            if (edge.ctx.leftToRight) { // look to the right                        
                previdx = nextidx;
                nextidx = searchEdgeCollectionRight(transcoder, edge.ctx, line, nextidx);
                found = checkStatus(edge.ctx, previdx, nextidx, nextidx === line.length);
            } else {// look to the left                        
                beforeidx = searchEdgeCollectionLeft(transcoder, edge.ctx, line, currentidx);
                found = checkStatus(edge.ctx, currentidx, beforeidx, beforeidx === 0);
            }
            if (!found) {
                match = "";
            } else {
                match = line.substring(n, idx);
            }
        } else {
            match = edge.out;
        }

        return match;
    };

    // move through line+idx until a character not in mi's edgeCollection is encountered.
    var searchEdgeCollectionRight = function(transcoder, ctx, line, idx) {
        var card, prefix;

        card = ctx.option;
        prefix = null;
        do {
            prefix = matchRight(ctx, transcoder.classes[ctx.category], line, idx);
            if (prefix !== null) {
                idx += prefix.length;
            }
            if (card === "1") {
                break;
            }
        } while (prefix !== null);

        return idx;
    };

    var searchEdgeCollectionLeft = function(transcoder, ctx, line, idx) {
        var card, prefix;

        card = ctx.option;
        prefix = null;
        do {
            prefix = matchLeft(ctx, transcoder.classes[ctx.category], line, idx);
            if (prefix !== null) {
                idx -= prefix.length;
            }
            if (card === "1") {
                break;
            }
        } while (prefix !== null);

        return idx;
    };

    var matchRight = function(ctx, classCollection, s, idx) {
        var len, key, obj;

        if (!ctx.maxEdgeLength) {
            computeMaxEdgeLength(ctx, classCollection);
        }
        // try longest edges first
        for (len = ctx.maxEdgeLength; len >= 1; len--) {
            if (idx + len > s.length) {   // len is too long, continue loop to get shorter length
                continue;
            }

            key = s.substring(idx, idx + len);
            if (classCollection[key] !== undefined) {
                return key;
            }
        }

        return null;
    };

    var matchLeft = function(ctx, classCollection, s, idx) {
        var len, key;

        if (!ctx.maxEdgeLength) {
            computeMaxEdgeLength(ctx, classCollection);
        }
        // try longest edges first
        for (len = ctx.maxEdgeLength; len >= 1; len--) {
            if (idx - len < 0) // len is too short, continue loop to get a shorter length
            {
                continue;
            }
            key = s.substring(idx - len, idx);
            if (classCollection[key] !== undefined) {
                return key;
            }
        }

        return null;
    };

    var computeMaxEdgeLength = function(ctx, classCollection) {
        var propname;
        ctx.maxEdgeLength = 0;
        for (propname in classCollection) {
            if (classCollection.hasOwnProperty(propname)) {
                if (propname.length > ctx.maxEdgeLength) {
                    ctx.maxEdgeLength = propname.length;
                }
            }
        }
    };

    var checkStatus = function(ctx, previdx, idx, atend) {
        var rc = false;

        if (ctx.eolOverride && atend && previdx === idx) // if eol is OK and nothing was matched because we are at the end (or beginning) of the line this overrides everything else.
        {
            rc = true;
        } else if (idx !== previdx) // something was matched
        {
            rc = ctx.exists;
        } else // no match found
        {
            if (!ctx.exists) // if it is not supposed to exist then mark it found
            {
                rc = true;
            } else if (ctx.option.equals("*")) // if it is okay for it not to exist, then mark it found
            {
                rc = true;       // don't break, there may be another
            } else // something must exist, so fail
            {
                rc = false;
            }
        }

        return rc;
    };

    var search = function(lines) {
        var query, i, len, line, percent, s, text, el, pattern, currentTarget, currentTargetHighlighted;

        if (!lines) {
            return;
        }

        if (currentAnswer === -1) {
            //query = document.getElementById("query").value;
            query = transcodeBuffer;
            currentTarget = transcodeText(lines, [query], false);
            currentTargetHighlighted = "<span style='background-color:silver'>" + currentTarget + "</span>";
            answers.length = 0;
            len = lines.length;
            for (i = 0; i < len; i++) {
                line = lines[i];
                if (line.indexOf(query) !== -1) {
                    answers.push(i);
                }
            }
            currentAnswer = 0;
        }

        if (currentAnswer !== answers.length) {
            if (prevMatchedLineObj) {
                prevMatchedLineObj.innerHTML = prevMatchedLineText;
            }
            el = document.getElementById("p" + answers[currentAnswer++]);
            if (currentAnswer === answers.length) {
                currentAnswer = 0;
            }
            prevMatchedLineObj = el;
            text = el.innerHTML;
            prevMatchedLineText = text;
            pattern = new RegExp(currentTarget, "g");
            text = text.replace(pattern, currentTargetHighlighted);
            el.innerHTML = text;
            api.scrollToY(el.offsetTop);
        }
    };

    var reset = function() {
        currentAnswer = -1;
    };

    var getKeyStroke = function(evt) {
        var ch, keyCode;

        //document.getElementById("debug").innerHTML = "" + evt.keyCode;
        if (evt.keyCode === 8) {
            transcodeBuffer = transcodeBuffer.substr(0, transcodeBuffer.length - 1);
        } else if (evt.keyCode !== 16) {
            keyCode = evt.keyCode;
            if (!evt.shiftKey) {
                keyCode += 32;          // make lower case
            }
            transcodeBuffer += String.fromCharCode(keyCode);
        }
    };

    var transcode = function(evt) {
        var s, obj;

        obj = document.getElementById("query");
        s = transcodeText([transcodeBuffer], {generateParagraphs: false});
        obj.value = s[0];
    };

    var getAlphabet = function() {
        var a, obj;

        a = [];
        for (obj in slpCategories) {			// obj iterates through SLP letters
            if (slpCategories.hasOwnProperty(obj)) {
                a.push(obj);
            }
        }
        return a;
    };

    return {
        inputIsSLP: inputIsSLP,
        getOtherBuffer: getOtherBuffer,
        addHandlersForDragging: addHandlersForDragging,
        getEchoTypingUsingDisplay: getEchoTypingUsingDisplay,
        setEchoTypingUsingDisplay: setEchoTypingUsingDisplay,
        addHandlersForTyping: addHandlersForTyping
    };
}();			// end of module T

