package org.sanskritlibrary.proxy;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.LineInfo;
import org.sanskritlibrary.ManuscriptInfo;
import org.sanskritlibrary.TextStructure;
import org.sanskritlibrary.Annotation;
import org.sanskritlibrary.servlet.SanskritLibraryServlet;
import org.sanskritlibrary.webservice.WebServices;

/**
 *
 * @author ralph
 */
public class LoadText extends SanskritLibraryServlet {
   
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        try {
            String text = request.getParameter("transcode");
            String anchor = request.getParameter("anchor");
            String htmlDirectory = (String) this.getServletContext().getInitParameter("dataDirectory") + "texts";
            String texttype = (String)request.getParameter("texttype");
            String textname = request.getParameter("text");         // e.g. MBh05UdyogaParvan
            if (text != null) {
                String srcEncoding = request.getParameter("srcEncoding");
                String tgtEncoding = request.getParameter("tgtEncoding");
                String transcodeDir = (String) this.getServletContext().getInitParameter("dataDirectory") + "transcoders";
                String result = WebServices.transformString(text, srcEncoding, tgtEncoding, transcodeDir);
                response.setContentType("text/plain; charset=UTF-8");
                PrintWriter pw = response.getWriter();
                pw.write(result);
            } else if (request.getParameter("test") != null) {
                String fn = String.format("%s/%s.slp", htmlDirectory, textname);
                File f = new File(fn);
                if (!f.exists()) {
                    fn = String.format("%s/%s.slp1", htmlDirectory, textname);                    
                }
                text = IOAssistance.readEntireFile(fn);
                this.writeResult(response, text, "text/text");
            } else if (anchor!=null) {
                String tsfn = String.format("%s/%s.bin", htmlDirectory, textname);
                String dir = request.getParameter("dir");           // directory where jpg files are
                String fn = String.format("%s/%s.roman", htmlDirectory, textname);
                text = IOAssistance.readEntireFile(fn);
                TextStructure ts = getTextStructure(request, tsfn);
                sendHTML(response, textname, text, ts, anchor);
            } else {
                if (textname.startsWith("texts/")) {
                    textname = textname.substring(6);
                }
                String fn = String.format("%s/%s/%s.json", htmlDirectory, texttype, textname);
                text = IOAssistance.readEntireFile(fn);
                if (texttype.equals("forManuscript")) {
                    
                }
                this.writeResult(response, text, "text/json");                
            }
        } catch (Exception ex) {
            sendMessage(response, ex.getMessage());
        }
    }

    private void sendHTML(HttpServletResponse response, String textname, String text, TextStructure ts, String anchor) throws IOException {
        String scriptTemplate = null;

        try {
            String textTemplateName = this.getServletContext().getRealPath("text2.template");
            scriptTemplate = IOAssistance.readEntireFile(textTemplateName);
        } catch (Exception ex) {
            throw new IOException(ex.getMessage());
        }

        String highlights = createHighlightsArray(ts.manuscriptDirectories.size(), ts.lineInfo);

        StringBuilder options = new StringBuilder();
        for (ManuscriptInfo mi : ts.manuscriptDirectories) {
            options.append(String.format("<option value='%s'>%s</option>\n", mi.directoryName, mi.directoryName));
        }
        response.setContentType("text/html; charset=UTF-8");
        String[] lines = text.split("\n");
        int numlines = lines.length;
        StringBuilder sb = new StringBuilder();
        for (int i=0; i<numlines; i++) {
            String line = lines[i];
            if (anchor != null && line.contains(anchor)) {
                sb.append(String.format("<a name='%s'/>\n", anchor));
            }
            sb.append(String.format("<p id='%d'>%s</p>\n", i, line));
        }

        PrintWriter pw = response.getWriter();
        pw.write(String.format(scriptTemplate, textname, textname, highlights, options.toString(), sb.toString()));
    }

    private String createHighlightsArray(int numManuscripts, ArrayList<LineInfo> lis) {
        StringBuilder sb = new StringBuilder();
        sb.append("[\n");
        String outerDelim = "\t";
        String delim = "";
        for (int i = 0; i < numManuscripts; i++) {
            delim = "";
            boolean firstTime = true;
            boolean exists = false;
            sb.append(String.format("%s[", outerDelim));
            outerDelim = ",\n\t";
            int count = 1;
            for (LineInfo li : lis) {
                Annotation a = li.annotations.get(i);
                if (firstTime) {
                    exists = a != null;
                    firstTime = false;
                    count = 1;
                } else {
                    if (exists) {
                        if (a != null) {
                            count++;
                        } else {
                            exists = false;
                            sb.append(String.format("%s%d", delim, count));
                            delim = ",";
                            count = 1;
                        }
                    } else {
                        if (a == null) {
                            count++;
                        } else {
                            exists = true;
                            sb.append(String.format("%s%d", delim, -count));
                            delim = ",";
                            count = 1;
                        }
                    }
                }
            }
            if (exists) {
                sb.append(String.format("%s%d", delim, count));
            } else {
                sb.append(String.format("%s%d", delim, -count));
            }
            sb.append("]");
        }

        sb.append("]");
        return sb.toString();
    }

    private void sendMessage(HttpServletResponse response, String msg) throws IOException {
        response.setContentType("text/html; charset=UTF-8");
        PrintWriter pw = response.getWriter();
        pw.write(msg);
    }

    /** 
     * Returns a short description of the servlet.
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "Short description";
    }// </editor-fold>

}
